from __future__ import annotations

import logging
import sys
import traceback
from typing import TYPE_CHECKING, Optional, override

import msgpack
import zmq

if TYPE_CHECKING:
    from ..logger import MsgpackOptions


class RelayHandler(logging.Handler):
    _sock: zmq.Socket | None

    def __init__(self, *, endpoint: str, msgpack_options: MsgpackOptions) -> None:
        super().__init__()
        self.endpoint = endpoint
        self.msgpack_options = msgpack_options
        self._zctx = zmq.Context[zmq.Socket]()
        # We should use PUSH-PULL socket pairs to avoid
        # lost of synchronization sentinel messages.
        if endpoint:
            self._sock = self._zctx.socket(zmq.PUSH)
            assert self._sock is not None
            self._sock.setsockopt(zmq.LINGER, 100)
            self._sock.connect(self.endpoint)
        else:
            self._sock = None

    def close(self) -> None:
        if self._sock is not None:
            self._sock.close()
        self._zctx.term()

    def _fallback(self, record: Optional[logging.LogRecord]) -> None:
        if record is None:
            return
        print(record.getMessage(), file=sys.stderr)

    @override
    def emit(self, record: Optional[logging.LogRecord]) -> None:
        if self._sock is None:
            self._fallback(record)
            return
        # record may be None to signal shutdown.
        if record:
            log_body = {
                "name": record.name,
                "pathname": record.pathname,
                "lineno": record.lineno,
                "msg": record.getMessage(),
                "levelno": record.levelno,
                "levelname": record.levelname,
            }
            if record.exc_info:
                log_body["exc_info"] = traceback.format_exception(*record.exc_info)
        else:
            log_body = None
        try:
            serialized_record = msgpack.packb(log_body, **self.msgpack_options["pack_opts"])
            self._sock.send(serialized_record)
        except zmq.ZMQError:
            self._fallback(record)



from __future__ import annotations

import logging
import pprint
import time
import traceback
from collections.abc import Sequence
from datetime import datetime
from types import TracebackType
from typing import Any, TypeAlias, cast

import coloredlogs
from pythonjsonlogger.json import JsonFormatter

_SysExcInfoType: TypeAlias = (
    tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None]
)


def format_exception(self, ei: Sequence[str] | _SysExcInfoType) -> str:
    match ei:
        case (str(), *_):
            # Already foramtted from the source process for ease of serialization
            s = "".join(cast(Sequence[str], ei))  # cast is required for mypy
        case (type(), BaseException(), _):
            # A live exc_info object from the current process
            s = "".join(traceback.format_exception(*ei))
        case _:
            s = "<exception-info-unavailable>"
    s = s.rstrip("\n")
    return s


class SerializedExceptionFormatter(logging.Formatter):
    def formatException(self, ei) -> str:
        return format_exception(self, ei)


class ConsoleFormatter(logging.Formatter):
    def formatException(self, ei) -> str:
        return format_exception(self, ei)

    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None) -> str:
        ct = self.converter(record.created)  # type: ignore
        if datefmt:
            datefmt = datefmt.replace("%f", f"{int(record.msecs):03d}")
            return time.strftime(datefmt, ct)
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            return f"{t}.{int(record.msecs):03d}"


class CustomJsonFormatter(JsonFormatter):
    def formatException(self, ei) -> str:
        return format_exception(self, ei)

    def add_fields(
        self,
        log_record: dict[str, Any],  # the manipulated entry object
        record: logging.LogRecord,  # the source log record
        message_dict: dict[str, Any],
    ) -> None:
        super().add_fields(log_record, record, message_dict)
        if not log_record.get("timestamp"):
            # this doesn't use record.created, so it is slightly off
            now = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            log_record["timestamp"] = now
        if loglevel := log_record.get("level"):
            log_record["level"] = loglevel.upper()
        else:
            log_record["level"] = record.levelname.upper()


class ColorizedFormatter(coloredlogs.ColoredFormatter):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        coloredlogs.logging.Formatter.formatException = format_exception


class pretty:
    """A simple object wrapper to pretty-format it when formatting the log record."""

    def __init__(self, obj: Any) -> None:
        self.obj = obj

    def __repr__(self) -> str:
        return pprint.pformat(self.obj)



from __future__ import annotations

import json
import logging
import socket
import ssl
from collections import OrderedDict
from datetime import datetime
from typing import override

import zmq

from ..exceptions import ConfigurationError


class LogstashHandler(logging.Handler):
    _sock: socket.socket | zmq.Socket | None
    _sslctx: ssl.SSLContext | None
    _zmqctx: zmq.Context | None

    def __init__(
        self,
        endpoint: tuple[str, int],
        protocol: str,
        *,
        ssl_enabled: bool = True,
        ssl_verify: bool = True,
        myhost: str | None = None,
    ) -> None:
        super().__init__()
        self._endpoint = endpoint
        self._protocol = protocol
        self._ssl_enabled = ssl_enabled
        self._ssl_verify = ssl_verify
        self._myhost = myhost
        self._sock = None
        self._sslctx = None
        self._zmqctx = None

    def _setup_transport(self) -> None:
        if self._sock is not None:
            return
        if self._protocol == "zmq.push":
            self._zmqctx = zmq.Context()
            zsock = self._zmqctx.socket(zmq.PUSH)
            zsock.setsockopt(zmq.LINGER, 50)
            zsock.setsockopt(zmq.SNDHWM, 20)
            zsock.connect(f"tcp://{self._endpoint[0]}:{self._endpoint[1]}")
            self._sock = zsock
        elif self._protocol == "zmq.pub":
            self._zmqctx = zmq.Context()
            zsock = self._zmqctx.socket(zmq.PUB)
            zsock.setsockopt(zmq.LINGER, 50)
            zsock.setsockopt(zmq.SNDHWM, 20)
            zsock.connect(f"tcp://{self._endpoint[0]}:{self._endpoint[1]}")
            self._sock = zsock
        elif self._protocol == "tcp":
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if self._ssl_enabled:
                self._sslctx = ssl.create_default_context()
                self._sslctx.minimum_version = ssl.TLSVersion.TLSv1_2
                if not self._ssl_verify:
                    self._sslctx.check_hostname = False
                    self._sslctx.verify_mode = ssl.CERT_NONE
                sock = self._sslctx.wrap_socket(sock, server_hostname=self._endpoint[0])
            sock.connect((self._endpoint[0], self._endpoint[1]))
            self._sock = sock
        elif self._protocol == "udp":
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.connect((self._endpoint[0], self._endpoint[1]))
            self._sock = sock
        else:
            raise ConfigurationError({
                "logging.LogstashHandler": f"unsupported protocol: {self._protocol}"
            })

    def cleanup(self) -> None:
        if self._sock:
            self._sock.close()
        self._sslctx = None
        if self._zmqctx:
            self._zmqctx.term()

    @override
    def emit(self, record: logging.LogRecord) -> None:
        self._setup_transport()
        tags: set[str] = set()

        # This log format follows logstash's event format.
        log = OrderedDict([
            ("@timestamp", datetime.now().isoformat()),
            ("@version", 1),
            ("host", self._myhost),
            ("logger", record.name),
            ("path", record.pathname),
            ("func", record.funcName),
            ("lineno", record.lineno),
            ("message", record.getMessage()),
            ("level", record.levelname),
            ("tags", list(tags)),
        ])
        if self._protocol.startswith("zmq"):
            assert isinstance(self._sock, zmq.Socket)
            self._sock.send_json(log)
        else:
            # TODO: reconnect if disconnected
            assert isinstance(self._sock, socket.socket)
            self._sock.sendall(json.dumps(log).encode("utf-8"))




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/logging:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
    ],
    'description': 'Backend.AI Logging Subsystem',
    'install_requires': (
        'coloredlogs~=15.0',
        'graypy==2.1.0',
        'msgpack~=1.1.0',
        'python-json-logger~=3.2.0',
        'pyzmq~=26.2',
        'trafaret~=2.1',
        'yarl~=1.13.1',
    ),
    'license': 'MIT',
    'long_description': """Backend.AI Logging Subsystem
============================

Package Structure
-----------------

* `ai.backend.logging`
  - `abc`: Abstract base classes
  - `logger`: The core logging facility
    - `Logger`: The standard multiprocess-friendly logger using `RelayHandler` based on ZeroMQ
    - `LocalLogger`: A minimalized console/file logger that does not require serialization via networks at all
  - `handler`: Collection of vendor-specific handler implementations
  - `formatter`: Collection of formatters
  - `types`: Definition of enums/types like `LogLevel`
  - `utils`: Brace-style message formatting adapters and other extras
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-logging',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.logging': (
            'VERSION',
            'py.typed',
        ),
    },
    'packages': (
        'ai.backend.logging',
        'ai.backend.logging.handler',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.12,<3.13',
    'url': 'https://github.com/lablup/backend.ai',
    'version': '25.3.0',
    'zip_safe': False,
})



from __future__ import annotations

from typing import Any, Mapping


class ConfigurationError(Exception):
    invalid_data: Mapping[str, Any]

    def __init__(self, invalid_data: Mapping[str, Any]) -> None:
        super().__init__(invalid_data)
        self.invalid_data = invalid_data
